@@ -5,6 +5,7 @@
 """
 
 import os
+import ast
 import math
 import asyncio
 import inspect
@@ -651,24 +652,54 @@ class rowParse:
         pass
 
 
+class LolPython(list):
+    """ because who needs to setattr lists right? """
+
+    def __setitem__(self, key, value):
+        super().__setitem__(key, value)
+        self._list[key] = value
+
+    def _asdict(self):
+        return {h:v for h, v in zip(self._header, self)}
+
+    def _raw(self):
+        return self._list
+
+
+def _sigh(header, row, orig_row=None):
+    if orig_row is None:
+        orig_row = row
+
+    out = LolPython(row)
+    out._list = orig_row
+    out._header = header
+    out._fields = header
+    for name, value in zip(header, out):
+        setattr(out, name, value)
+
+    return out
+
+
 class byCol:
     # FIXME this class +is+ was unpickleable, lol python
     def __init__(self, rows, header=None, to_index=tuple()):
         # what I think is a pickleable version
         if header is None:  # FIXME non None header might have bad names?
-            orig_header = [str(c) for c in rows[0]]  # normalize all to string for safety
-            header = [python_identifier(c) for i, c in enumerate(orig_header)]
+            _oh = [str(c) for c in rows[0]]
+            header = [python_identifier(c) for i, c in enumerate(_oh)]
             #changes = {new:old for old, new in zip(rows[0], header) if old != new}
+            orig_header = _sigh(header, _oh)  # normalize all to string for safety
             rows = rows[1:]
         else:
-            orig_header = header
+            orig_header = _sigh(header, header)
 
         # normalize row lenght  # FIXME account for rows longer than header
-        rows = [row + [None] * (len(header) - len(row)) for row in rows]
+        rows = [_sigh(header, row + [None] * (len(header) - len(row)), row) for row in rows]
+        #setattr(self, name, LOL_PYTHON(header, row))
 
         # so apparently using namedtuple dynamically breaks pickle LOL PYTHON LOL
         self.orig_header = orig_header
-        self.header = header
+        self.header = _sigh(header, header)
         self.rows = rows
         self.__indexes = {}
         for col_name in to_index:
@@ -732,8 +763,9 @@ class byCol:
         for col in self.header:
             yield [col, *getattr(self, col)]
 
-    def searchIndex(self, index, value):
-        return self.__indexes[index][value]
+    def searchIndex(self, index, value, raw=False):
+        row = self.__indexes[index][value]
+        return row._raw() if raw else row
 
     def __getitem__(self, key):
         return list(getattr(self, key))
@@ -792,3 +824,133 @@ class _TermColors:
 
 
 TermColors = _TermColors()
+
+
+## ast printing
+
+def findAssignToName(name, body, reverse=False, col_offset=0):
+    eb = enumerate(body)
+    gen = reversed(list(eb)) if reverse else eb
+    return [(i, n) for i, n in gen
+            if n.col_offset <= col_offset and
+            isinstance(n, ast.Assign) and
+            isinstance(n.targets[0], ast.Name) and
+            n.targets[0].id == name]
+
+
+def findImportName(name, body, reverse=False, col_offset=0):
+    eb = enumerate(body)
+    gen = reversed(list(eb)) if reverse else eb
+    for i, node in gen:
+        if node.col_offset <= col_offset:
+            if not (isinstance(node, ast.Import) or
+                    isinstance(node, ast.ImportFrom)):
+                continue
+
+            if name in [n.asname if n.asname
+                        else n.name for n in node.names]:
+                return [(i, node)]
+
+
+def findName(name, body, reverse=False):
+    out = findAssignToName(name, body, reverse=reverse)
+    if not out:
+        out = findImportName(name, body, reverse=reverse)
+
+    return out
+
+
+def asStr(astNode, prior=tuple(), dereference=False):
+    """ having to explicitly pass prior to all recursive calls:
+        racket: am I a joke to you? sigh continuation markes talk
+        is just timely enough to realize how useful dynamic extent is
+        and just how much we miss it in cases like this >_< """
+
+    if isinstance(astNode, ast.Str):
+        return repr(astNode.s)
+    elif isinstance(astNode, ast.FormattedValue):
+        return asStr(astNode.value, prior, dereference=True)
+    elif isinstance(astNode, ast.JoinedStr):
+        around = '"""'
+        return (around +
+                ''.join([ast.literal_eval(asStr(v, prior))
+                         if isinstance(v, ast.Str) else asStr(v, prior)
+                         for v in astNode.values]) +
+                around)
+    elif isinstance(astNode, ast.Name):
+        if prior:
+            # have to reverse to find the latest setting of the value
+            # not the first since python allows rebinding
+            maybe_name = findName(astNode.id, prior, reverse=True)
+            if maybe_name:
+                index, node = maybe_name[0]
+                s = asStr(node, prior[:index],
+                          dereference=astNode.id if dereference else False)
+                return s
+            else:
+                raise NotImplementedError(f'found no assignment to {astNode.id}')
+
+        return astNode.id
+    elif isinstance(astNode, ast.Assign):
+        value = asStr(astNode.value, prior, dereference=dereference)
+        if dereference:
+            return value
+        else:
+            return ', '.join([asStr(n, dereference=dereference)
+                              for n in astNode.targets]) + ' = ' + value
+    elif isinstance(astNode, ast.BinOp):
+        return (asStr(astNode.left, prior, dereference) +
+                ' ' + asStr(astNode.op, prior, dereference) + ' ' +
+                asStr(astNode.right, prior, dereference))
+    elif isinstance(astNode, ast.Add):
+        return '+'
+    elif isinstance(astNode, ast.Div):
+        return '/'
+    elif isinstance(astNode, ast.Call):
+        return (asStr(astNode.func, prior, dereference) +
+                '(' +
+                ', '.join(asStr(a, prior, dereference) for a in astNode.args) +
+                ')')
+    elif isinstance(astNode, ast.Attribute):
+        return asStr(astNode.value, prior, dereference) + '.' + astNode.attr
+    elif isinstance(astNode, ast.Dict):
+        return ('{' +
+                ', '.join([(asStr(key, prior, dereference) +
+                            ': ' + asStr(value, prior, dereference))
+                           for key, value in zip(astNode.keys, astNode.values)]) +
+                '}')
+    elif isinstance(astNode, ast.Lambda):
+        return ('lambda ' +
+                ', '.join([asStr(a, prior, dereference) for a in astNode.args.args]) +
+                ': ' +
+                asStr(astNode.body, prior, dereference))
+    elif isinstance(astNode, ast.Expr):
+        return asStr(astNode.value, prior, dereference)
+    elif isinstance(astNode, ast.Module):
+        return '\n'.join([asStr(b, prior, dereference) for b in astNode.body])
+    elif isinstance(astNode, ast.arg):
+        return astNode.arg
+    elif isinstance(astNode, ast.Import):
+        return ('import ' + ', '.join([asStr(n) for n in astNode.names]))
+    elif isinstance(astNode, ast.ImportFrom):
+        _names = [asStr(n, dereference=dereference) for n in astNode.names]
+        names = [n for n in _names if n]  # filter the dereferenced case
+        if dereference:
+            name = names[0]  # a, b = imported_name1, imported_name2 not an issue, resolved independently
+            return astNode.module + '.' + name
+        else:
+            return ('from ' + astNode.module + ' ' +
+                    'import ' + ', '.join(names))
+    elif isinstance(astNode, ast.alias):
+        if dereference:
+            if (astNode.asname == dereference or
+                not astNode.asname and astNode.name == dereference):
+                # this deals with the threat of
+                # from asdf import name as dereference, dereference as name
+                return astNode.name
+        else:
+            return astNode.name + ((' as ' + astNode.asname)
+                                   if astNode.asname else
+                                   '')
+    else:
+        raise NotImplementedError(f'{astNode}')
